;==============================================================================

;----------------------------------------------------------------------------
; This structure, the loadall_table in the data section, and the code below
; that is labeled as such adapted from the 386 LOADALL SetESLimit example
; in Geoff Chappell's book DOS Internals, 1994, Addison-Wesley,
; ISBN 0-201-60835-9
;----------------------------------------------------------------------------

;----------------------------------------------------------------
; Descriptor cache format for the 80386 machine state structure:
;----------------------------------------------------------------

CACHE STRUCT
    unknown_00h   db ?
    access_rights db ?
    unknown_02h   db ?
    unknown_03h   db ?
    base          dd ?
    limit         dd ?
CACHE ENDS

;==============================================================================

;------------------------------
; Macros borrowed from MASM32:
;------------------------------

    reparg MACRO arg
      LOCAL nustr
      quot SUBSTR <arg>, 1, 1
      IFIDN quot,<">          ;; if 1st char = "
        .data
          nustr db arg, 0     ;; write arg to .DATA section
        .code
        EXITM <ADDR nustr>    ;; append name to ADDR operator
      ELSE
        EXITM <arg>           ;; else return arg
      ENDIF
    ENDM

    chr$ MACRO any_text:VARARG
      LOCAL txtname
      .data
        txtname db any_text,0
      .code
      EXITM <OFFSET txtname>
    ENDM

    print MACRO arg1:REQ,varname:VARARG ;; display zero terminated string
      invoke StdOut,reparg(arg1)
      IFNB <varname>
        invoke StdOut,chr$(varname)
      ENDIF
    ENDM

    hex$ MACRO DDvalue
      LOCAL rvstring
      .data
        rvstring db 12 dup (0)
        ;align 4                ; invalid for 16-bit code
      .code
      invoke dw2hex,DDvalue,ADDR rvstring
      EXITM <ADDR rvstring>
    ENDM

;==============================================================================

.model small, c
.386p

;----------------------------------------------------
; The following were used to test the invalid opcode
; exception code on my P3 with the SSE2 instruction:
; addpd xmm1, xmm2
;----------------------------------------------------
;.686p
;.MMX
;.XMM

;==============================================================================
.stack
;==============================================================================
.data

    ;--------------------------------------------------------------------
    ; Statically allocated machine state structure for the 80386 LOADALL
    ; instruction - observe that all segments are given access rights
    ; of 92h:
    ;--------------------------------------------------------------------

    loadall_table       LABEL BYTE
    loadall_cr0         dd ?
    loadall_eflags      dd ?
    loadall_eip         dd OFFSET after_loadall
    loadall_edi         dd ?
    loadall_esi         dd ?
    loadall_ebp         dd ?
    loadall_esp         dd ?
    loadall_ebx         dd ?
    loadall_edx         dd ?
    loadall_ecx         dd ?
    loadall_eax         dd ?
    loadall_dr6         dd ?
    loadall_dr7         dd ?
    loadall_tr          dd ?
    loadall_ldtr        dd ?
    loadall_gs          dd ?
    loadall_fs          dd ?
    loadall_ds          dd ?
    loadall_ss          dd ?
    loadall_cs          dd ?
    loadall_es          dd ?

    loadall_tr_cache    CACHE <>
    loadall_idtr_cache  CACHE <>
    loadall_gdtr_cache  CACHE <>
    loadall_ldtr_cache  CACHE <>
    loadall_gs_cache    CACHE <00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>
    loadall_fs_cache    CACHE <00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>
    loadall_ds_cache    CACHE <00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>
    loadall_ss_cache    CACHE <00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>
    loadall_cs_cache    CACHE <00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>
    loadall_es_cache    CACHE <00h, 92h, 00h, 00h, 00000000h, 0000FFFFh>

    scratch       FWORD 0
    prev_isr_06   dd    0
    hexChars      db    "0123456789ABCDEF"

.code
;==============================================================================

szlen proc c uses bx npsz:WORD
    mov bx, npsz
  @@:
    cmp BYTE PTR [bx], 0
    jz  @F
    inc bx
    jmp @B
  @@:
    mov ax, bx
    sub ax, npsz
    ret
szlen endp

HSTDOUT EQU 1

stdout proc c uses ax bx cx dx npsz:WORD
    invoke szlen, npsz
    mov cx, ax
    mov bx, HSTDOUT
    mov dx, npsz
    mov ah, 40h
    int 21h
    ret
stdout endp

dw2hex proc c uses ax bx dx di src:DWORD, npdest:WORD
    mov edx, src
    mov di, npdest
    add di, 7
    mov BYTE PTR [di+1], 0
  @@:
    mov ebx, edx
    and ebx, 0Fh
    mov al, hexChars[ebx]
    mov [di], al
    shr edx, 4
    dec di
    cmp di, npdest
    jnb @B
    ret
dw2hex endp

;==============================================================================

  isr_06:

    pushad
    print "INVALID OPCODE",13,10
    popad
    pop ax
    mov ax, invalid_opcode_bypass
    push ax
    iret

;==============================================================================

.startup

    mov ax, 3506h
    int 21h
    mov WORD PTR prev_isr_06,bx
    mov WORD PTR prev_isr_06[2],es
    mov ax, 2506h
    push ds
    push cs
    pop ds
    mov dx, OFFSET isr_06
    int 21h
    pop ds

    ;---------------------------
    ; Begin DOS Internals code.
    ;---------------------------

;   Preserve registers important to system control (and whose values can be
;   determined in real mode) - the GDTR and IDTR table registers, CR0 and
;   the two debugging registers DR6 and DR7.

    sgdt  [scratch]
    mov ax,word ptr [scratch]
    mov word ptr [loadall_gdtr_cache.limit],ax
    mov eax,dword ptr [scratch + 02h]
    mov [loadall_gdtr_cache.base],eax

    sidt  [scratch]
    mov ax,word ptr [scratch]
    mov word ptr [loadall_idtr_cache.limit],ax
    mov eax,dword ptr [scratch + 02h]
    mov [loadall_idtr_cache.base],eax

    mov eax,cr0
    mov [loadall_cr0],eax
    mov eax,dr6
    mov [loadall_dr6],eax
    mov eax,dr7
    mov [loadall_dr7],eax

    pushfd
    pop [loadall_eflags]

;   Make sure that all registers needed for access to the stack will retain
;   their values.  Visible and invisible components of the SS register must
;   be prepared - note that a 64KB limit for SS has been set statically.

    mov [loadall_ebp],ebp
    mov [loadall_esp],esp

    mov ax,ss
    mov word ptr [loadall_ss],ax
    movzx eax,ax
    shl eax,4
    mov [loadall_ss_cache.base],eax

;   Similarly, make sure that CS:IP will address the instruction immediately
;   after LOADALL.  The value for the IP register is set statically.

    mov ax,cs
    mov word ptr [loadall_cs],ax
    movzx eax,ax
    shl eax,4
    mov [loadall_cs_cache.base],eax


;   The LOADALL instruction expects the machine state structure at ES:DI.
;   Immediately after the LOADALL instruction, all registers (including the
;   invisible components of segment registers) have values taken from the
;   machine state structure.

    push ds
    pop es
    mov di,OFFSET loadall_table

    ;-------------------------
    ; End DOS Internals code.
    ;-------------------------

    db 0fh, 07h

  after_loadall:

    mov ax, @data
    mov ds, ax

    print "LOADALL OK",13,10

  invalid_opcode_bypass:

    push ds
    lds dx, prev_isr_06
    mov ax, 2506h
    int 21h
    pop ds

    print "Press any key to exit...",13,10

    mov ah, 0
    int 16h

.exit

;==============================================================================

end